import 'dart:async';
import 'dart:io';

import 'package:audioplayers/audioplayers.dart';
import 'package:base_utils/share_preference/share_preference_manager.dart';
import 'package:base_utils/utils/logging_utils.dart';
import 'package:flutter/services.dart';
import 'package:path_provider/path_provider.dart';

enum PlayMode { Once, Repeat }

class SoundManager {
  static AudioPlayer audioPlayer = new AudioPlayer();

  static bool _enabled = true;

  static bool _stopped = true;

  static bool get isEnabled {
    return SharePreferenceManager.getPrefsBool(ConstantsSharePref.DB_SOUND);
  }

  static final _filesMap = Map<String, File>();

  static var duration = Duration();

  static final interval = Duration(milliseconds: 100);

  static Timer timer;

  static bool muted = false;

  static const TAG = "SoundManager";

  static void setEnabled(bool enabled, {bool fromSystem = true}) {
    assert(enabled != null);

    log("$TAG: setEnabled: $enabled, fromSystem:$fromSystem");

    if (fromSystem) {
      SharePreferenceManager.setPrefs(
          ConstantsSharePref.DB_SOUND, Type.BOOL, enabled);
      muted = !enabled;
    } else {
      if (enabled && muted) {
        return;
      }
    }

    _enabled = enabled;


    if (!enabled) {
      audioPlayer.pause();
    } else {
      log("$TAG: Considering playing audio");
      if (!_stopped) {
        log("$TAG: Not stopped -> play audio");
        audioPlayer.resume();
        audioPlayer.seek(duration);
      }
    }
  }

  static Future init(List<String> fileNames, {String package}) async {
    audioPlayer.setReleaseMode(ReleaseMode.STOP);
    print(
        '~~~~> Ahihi: SharePreferenceManager.getPrefsBool(ConstantsSharePref.DB_SOUND)${SharePreferenceManager
            .getPrefsBool(ConstantsSharePref.DB_SOUND)}');
    fileNames.forEach((localFileName) async {
      await load(localFileName, package: package);
    });
  }

  static Future<File> load(localFileName, {String package}) async {
    final dir = await getApplicationDocumentsDirectory();
    final file = new File("${dir.path}/$localFileName");
    final assetPath = generateAssetPath(localFileName, package);
    if (!(await file.exists())) {
      final soundData = await rootBundle.load(assetPath);
      final bytes = soundData.buffer.asUint8List();
      await file.writeAsBytes(bytes, flush: true);
    }
    _filesMap[assetPath] = file;
    return file;
  }

  static String generateAssetPath(String localFileName, String package) =>
      "packages/$package/sounds/$localFileName";

  static Future play(String localFileName,
      {String package, PlayMode mode = PlayMode.Once}) async {
    log("$TAG: play: $localFileName, ${mode.toString()}");

    final assetPath = generateAssetPath(localFileName, package);

    if (!_filesMap.containsKey(assetPath)) {
      log("$TAG: loading: $localFileName");
      await load(localFileName, package: package);
      log("$TAG: done loading: $localFileName");
    }

    duration = Duration();

    _stopped = false;

    cancelTimer();

    timer = Timer.periodic(interval, (timer) {
      duration += interval;
    });

    audioPlayer.completionHandler = () async {
      cancelTimer();

      _stopped = false;

      if (mode == PlayMode.Repeat) {
        log("$TAG: repeating...");
        play(localFileName, package: package, mode: mode);
      } else {
        log("$TAG: stop after played once...");
        _stopped = true;
      }
    };

    audioPlayer.stop();

    await audioPlayer.play(_filesMap[assetPath].path, isLocal: true);
    log("$TAG: after called play.");

    if (!SharePreferenceManager.getPrefsBool(ConstantsSharePref.DB_SOUND)) {
      Future.delayed(interval, () {
        log("$TAG: is disabled, pause...");
        audioPlayer.pause();
      });
    }
  }

  static void stop() {
    audioPlayer.stop();
    _stopped = true;
    cancelTimer();
  }

  static void cancelTimer() {
    if (timer != null) {
      timer.cancel();
      timer = null;
    }
  }

  static void release() {
    audioPlayer.release();
    _enabled = true;
  }
}
